"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DatabricksQuery = void 0;
const ramda_1 = __importDefault(require("ramda"));
const schema_compiler_1 = require("@cubejs-backend/schema-compiler");
const shared_1 = require("@cubejs-backend/shared");
const GRANULARITY_TO_INTERVAL = {
    day: 'day',
    week: 'week',
    hour: 'hour',
    minute: 'minute',
    second: 'second',
    month: 'month',
    quarter: 'quarter',
    year: 'year'
};
class DatabricksFilter extends schema_compiler_1.BaseFilter {
    likeIgnoreCase(column, not, param, type) {
        const p = (!type || type === 'contains' || type === 'ends') ? '%' : '';
        const s = (!type || type === 'contains' || type === 'starts') ? '%' : '';
        return `LOWER(${column})${not ? ' NOT' : ''} LIKE CONCAT('${p}', LOWER(${this.allocateParam(param)}), '${s}')`;
    }
}
class DatabricksQuery extends schema_compiler_1.BaseQuery {
    newFilter(filter) {
        return new DatabricksFilter(this, filter);
    }
    castToString(sql) {
        return `CAST(${sql} as STRING)`;
    }
    hllInit(sql) {
        return `hll_sketch_agg(${sql})`;
    }
    hllMerge(sql) {
        return `hll_union_agg(${sql})`;
    }
    hllCardinality(sql) {
        return `hll_sketch_estimate(${sql})`;
    }
    hllCardinalityMerge(sql) {
        return `hll_sketch_estimate(hll_union_agg(${sql}))`;
    }
    countDistinctApprox(sql) {
        return `approx_count_distinct(${sql})`;
    }
    convertTz(field) {
        return `from_utc_timestamp(${field}, '${this.timezone}')`;
    }
    timeStampCast(value) {
        return `from_utc_timestamp(replace(replace(${value}, 'T', ' '), 'Z', ''), 'UTC')`;
    }
    dateTimeCast(value) {
        return `from_utc_timestamp(${value}, 'UTC')`; // TODO
    }
    subtractInterval(date, interval) {
        const intervalParsed = (0, shared_1.parseSqlInterval)(interval);
        let res = date;
        for (const [key, value] of Object.entries(intervalParsed)) {
            res = `(${res} - INTERVAL '${value}' ${key})`;
        }
        return res;
    }
    addInterval(date, interval) {
        const intervalParsed = (0, shared_1.parseSqlInterval)(interval);
        let res = date;
        for (const [key, value] of Object.entries(intervalParsed)) {
            res = `(${res} + INTERVAL '${value}' ${key})`;
        }
        return res;
    }
    timeGroupedColumn(granularity, dimension) {
        return `date_trunc('${GRANULARITY_TO_INTERVAL[granularity]}', ${dimension})`;
    }
    /**
     * Returns sql for source expression floored to timestamps aligned with
     * intervals relative to origin timestamp point.
     */
    dateBin(interval, source, origin) {
        const [intervalFormatted, timeUnit] = this.formatInterval(interval);
        const beginOfTime = this.dateTimeCast('\'1970-01-01T00:00:00\'');
        return `${this.timeStampCast(`'${origin}'`)} + INTERVAL ${intervalFormatted} *
      floor(
        date_diff(${timeUnit}, ${this.timeStampCast(`'${origin}'`)}, ${source}) /
        date_diff(${timeUnit}, ${beginOfTime}, ${beginOfTime} + INTERVAL ${intervalFormatted})
      )`;
    }
    /**
     * The input interval with (possible) plural units, like "2 years", "3 months", "4 weeks", "5 days"...
     * will be converted to Databricks dialect.
     * @see https://docs.databricks.com/en/sql/language-manual/data-types/interval-type.html
     * It returns a tuple of (formatted interval, timeUnit to use in datediff functions)
     */
    formatInterval(interval) {
        const intervalParsed = (0, shared_1.parseSqlInterval)(interval);
        const intKeys = Object.keys(intervalParsed).length;
        if (intervalParsed.year && intKeys === 1) {
            return [`'${intervalParsed.year}' YEAR`, 'YEAR'];
        }
        else if (intervalParsed.year && intervalParsed.month && intKeys === 2) {
            return [`'${intervalParsed.year}-${intervalParsed.month}' YEAR TO MONTH`, 'MONTH'];
        }
        else if (intervalParsed.month && intKeys === 1) {
            return [`'${intervalParsed.month}' MONTH`, 'MONTH'];
        }
        else if (intervalParsed.day && intKeys === 1) {
            return [`'${intervalParsed.day}' DAY`, 'DAY'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intKeys === 2) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}' DAY TO HOUR`, 'HOUR'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intKeys === 3) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}' DAY TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 4) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' DAY TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.hour && intervalParsed.minute && intKeys === 2) {
            return [`'${intervalParsed.hour}:${intervalParsed.minute}' HOUR TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 3) {
            return [`'${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' HOUR TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.minute && intervalParsed.second && intKeys === 2) {
            return [`'${intervalParsed.minute}:${intervalParsed.second}' MINUTE TO SECOND`, 'SECOND'];
        }
        // No need to support microseconds.
        throw new Error(`Cannot transform interval expression "${interval}" to Databricks dialect`);
    }
    escapeColumnName(name) {
        return `\`${name}\``;
    }
    getFieldIndex(id) {
        const dimension = this.dimensionsForSelect().find((d) => d.dimension === id);
        if (dimension) {
            return super.getFieldIndex(id);
        }
        return this.escapeColumnName(this.aliasName(id, false));
    }
    unixTimestampSql() {
        return 'unix_timestamp()';
    }
    orderHashToString(hash) {
        if (!hash || !hash.id) {
            return null;
        }
        const fieldIndex = this.getFieldIndex(hash.id);
        if (fieldIndex === null) {
            return null;
        }
        const dimensionsForSelect = this.dimensionsForSelect();
        const dimensionColumns = ramda_1.default.flatten(dimensionsForSelect.map((s) => s.selectColumns() && s.aliasName()))
            .filter(s => !!s);
        if (dimensionColumns.length) {
            const direction = hash.desc ? 'DESC' : 'ASC';
            return `${fieldIndex} ${direction}`;
        }
        return null;
    }
    defaultRefreshKeyRenewalThreshold() {
        return 120;
    }
    sqlTemplates() {
        const templates = super.sqlTemplates();
        templates.functions.DATETRUNC = 'DATE_TRUNC({{ args_concat }})';
        templates.functions.DATEPART = 'DATE_PART({{ args_concat }})';
        templates.functions.BTRIM = 'TRIM({% if args[1] is defined %}{{ args[1] }} FROM {% endif %}{{ args[0] }})';
        templates.functions.LTRIM = 'LTRIM({{ args|reverse|join(", ") }})';
        templates.functions.RTRIM = 'RTRIM({{ args|reverse|join(", ") }})';
        templates.functions.DATEDIFF = 'DATEDIFF({{ date_part }}, DATE_TRUNC(\'{{ date_part }}\', {{ args[1] }}), DATE_TRUNC(\'{{ date_part }}\', {{ args[2] }}))';
        templates.functions.LEAST = 'LEAST({{ args_concat }})';
        templates.functions.GREATEST = 'GREATEST({{ args_concat }})';
        templates.functions.TRUNC = 'CASE WHEN ({{ args[0] }}) >= 0 THEN FLOOR({{ args_concat }}) ELSE CEIL({{ args_concat }}) END';
        templates.expressions.timestamp_literal = 'from_utc_timestamp(\'{{ value }}\', \'UTC\')';
        templates.expressions.extract = 'EXTRACT({{ date_part }} FROM {{ expr }})';
        templates.quotes.identifiers = '`';
        templates.quotes.escape = '``';
        // TODO: Databricks has `TIMESTAMP_NTZ` with logic similar to Pg's `TIMESTAMP`
        // but that requires Runtime 13.3+. Should this be enabled?
        // templates.types.timestamp = 'TIMESTAMP_NTZ';
        delete templates.types.time;
        // Databricks intervals have a YearMonth or DayTime type variants, but no universal type
        delete templates.types.interval;
        return templates;
    }
}
exports.DatabricksQuery = DatabricksQuery;
//# sourceMappingURL=DatabricksQuery.js.map